/*
 * This file is part of the MDR library.
 * Copyright (C) 2013-2024 Mateusz Viste
 */

#include <mdr\video.h> /* video_detectega() */

#include "mdr\cout.h"  /* include self headers for control */


/* a few global variables used to maintain some stateful data in memory */
static unsigned char far *vmem; /* video memory pointer (beginning of page 0) */
static unsigned char term_width = 0, term_height = 0;
static unsigned char savedattr = 0; /* used to restore the same video mode before quitting */
static unsigned char cursor_x = 0, cursor_y = 0; /* current cursor coordinates (cached) */
static unsigned short cursor_shape = 0; /* remember the cursor's shape */


/* I get the current video mode first because according to the RBIL:
 * "BUG: AMI 386 BIOS and AST Premier 386 BIOS will lock up the system if
 *       AL is not equal to the current video mode" */
static void cursor_set(unsigned short shape);
#pragma aux cursor_set = \
"mov ah, 0x0f" /* get cur video mode */ \
"int 0x10"     /*  AL=video mode ; AH = num of columns ; BH = active page */ \
"mov ah, 1" \
"int 0x10" \
parm [cx] \
modify [ax bh];


static void goto_xy(unsigned char row, unsigned char col);
#pragma aux goto_xy = \
"mov ah, 0x02" \
"xor bh, bh" \
"int 0x10" \
parm [dh] [dl] \
modify [ah bh];


static void internal_cls(unsigned char colattr, unsigned char termh, unsigned char termw);
#pragma aux internal_cls = \
"mov ax, 0x0600"       /* Scroll window up - entire window */ \
"xor cx, cx"           /* Upper left */ \
"dec dh" \
"dec dl" \
"int 0x10" \
parm [bh] [dh] [dl] \
modify [ax cx bp];     /* flagging BP as modified because according to RBIL:
                          "Some implementations (including the original IBM PC)
                          have a bug which destroys BP." */


void mdr_cout_cursor_hide(void) {
  /* shape 0x2020 is suggested by TechHelp:
   * http://www.techhelpmanual.com/116-int_10h_01h__set_cursor_shape___size.html */
  cursor_set(0x2020); /* hide the cursor (start scanline = 0x20 / end scanline = 0x20 */
}


void mdr_cout_cursor_show(void) {
  cursor_set(cursor_shape); /* unhide the cursor */
}


/* get current attribute value under cursor and returns it */
int mdr_cout_getcurattr(void) {
  unsigned char resah = 0;
  _asm {
    mov ah, 0x08      /* Read character and attribute at cursor position */
    xor bh, bh        /* display page */
    int 0x10
    mov resah, ah
  }
  return(resah);
}


/* gets cursor's position on screen (row, column) and shape */
void mdr_cursor_getinfo(unsigned char *column, unsigned char *row, unsigned short *shape) {
  unsigned char regdl = 0, regdh = 0;
  unsigned short regcx = 0;
  _asm {
    push bx
    push cx
    push dx
    mov ah, 3      /* get cursor position and shape */
    xor bh, bh     /* page number (pretty much always 0) */
    int 0x10
    mov regdl, dl
    mov regdh, dh
    mov regcx, cx
    pop dx
    pop cx
    pop bx
  }
  if (column) *column = regdl; /* column */
  if (row) *row = regdh;       /* row */
  if (shape) *shape = regcx;   /* start scan line (ch) + end scan line (cl) */
}


/* inits the subsystem, fills arguments with:
 * w = screen width
 * h = screen height
 * Any of these arguments may be passed as NULL
 * Returns a color flag (0=mono, 1=color) */
unsigned char mdr_cout_init(unsigned char *w, unsigned char *h) {
  unsigned char color_flag;
  savedattr = mdr_cout_getcurattr(); /* there we save current color attributes for later restoration */
  mdr_cout_getconprops(&term_width, &term_height, &color_flag);
  mdr_cursor_getinfo(&cursor_x, &cursor_y, &cursor_shape); /* save the current cursor's position and shape */
  if (w) *w = term_width;
  if (h) *h = term_height;
  return(color_flag);
}


void mdr_cout_close(void) {
  mdr_cout_cls(savedattr); /* clears the screen and restore the original colors of the shell */
}


void mdr_cout_locate(unsigned char row, unsigned char column) {
  /* cache new cursor coordinates*/
  cursor_y = row;
  cursor_x = column;
  /* and apply them */
  goto_xy(row, column);
}


/* print a single character on screen */
void mdr_cout_char(unsigned char y, unsigned char x, char c, unsigned char attr) {
  unsigned char far *p;
  p = vmem + ((y * term_width + x) << 1);
  *p++ = c;
  *p = attr;
}


/* print a single character on screen, repeated count times */
void mdr_cout_char_rep(unsigned char y, unsigned char x, char c, unsigned char attr, unsigned char count) {
  unsigned char far *p;
  p = vmem + ((y * term_width + x) << 1);
  while (count--) {
    *p++ = c;
    *p++ = attr;
  }
}


/* print a nul-terminated string on screen, up to maxlen characters
 * returns the number of characters actually displayed */
unsigned char mdr_cout_str(unsigned char y, unsigned char x, const char *s, unsigned char attr, unsigned char maxlen) {
  unsigned char far *p;
  unsigned char n = 0;
  p = vmem + ((y * term_width + x) << 1);
  while ((*s != 0) && (maxlen != n)) {
    *p++ = *s++;
    *p++ = attr;
    n++;
  }
  return(n);
}


/* clears screen, filling it with a single color attribute */
void mdr_cout_cls(unsigned char colattr) {
  mdr_cout_locate(0, 0);
  internal_cls(colattr, term_height, term_width);
}


void mdr_cout_getconprops(unsigned char *termwidth, unsigned char *termheight, unsigned char *colorflag) {
  unsigned char regal = 0, regah = 0;

  _asm {
    mov ah, 0x0f       /* get current video mode */
    int 0x10
    mov regal, al
    mov regah, ah
  }

  *termwidth = regah; /* int10,F provides number of columns in AH */
  /* if EGA+ detected, fetch number of rows, otherwise (CGA...) assume 25 */
  if (video_detectega() == 0) {
    /* the byte at 0040:0084 identifies the maximum row value, on EGA/VGA BIOSes */
    *termheight = (*(unsigned char far *) 0x00400084l) + 1; /* 0040:0084 */
  } else {
    *termheight = 25;
  }
  /* select the most appropriate color scheme */
  if ((regal == 0) || (regal == 2) || (regal == 7)) { /* mono modes */
    *colorflag = 0;
  } else { /* else we are in color */
    *colorflag = 1;
  }
  /* select the correct VRAM address */
  if (regal == 7) { /* MDA/HERC mode */
    vmem = (unsigned char far *) 0xB0000000l; /* B000:0000 video memory addess */
  } else {
    vmem = (unsigned char far *) 0xB8000000l; /* B800:0000 video memory address */
  }
}
