/*
 * CTRL+C handling
 *
 * This file is part of the Mateusz' DOS Routines (MDR): http://mdr.osdn.io
 * Published under the terms of the MIT License, as stated below.
 *
 * Copyright (C) 2014-2023 Mateusz Viste
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#include <mdr/dos.h> /* include self */

/* sets up the CTRL+C handler for the running program to a no-op - in other
 * words, after this call DOS will no longer abort the program on CTRL+C.
 * this is only valid for the duration of the program because DOS will restore
 * the original handler after the program exits.
 *
 * an alternative is mdr_dos_ctrlc_off(), but this does not inhibit the
 * handler, it sets DOS to not react to CTRL+C in the first place, and this
 * setting stays active after the program quits so the program should remember
 * to restore the original setting before quitting. */
void mdr_dos_ctrlc_inhibit(void) {
  static unsigned char op_iret = 0xCF;

  /* disable the CTRL+C handler by setting up the INT 23h interrupt vector to
   * an iret (opcode 0xCF) */

  _asm {
    mov ax, 0x2523 /* AH=25h = DOS 1+ - set interrupt vector ; AL=23h int23 */
    mov dx, offset [op_iret]
    int 0x21
  }
}
