/*
 * SELFEXE - returns the full path and filename of the running program.
 *
 * This file is part of the Mateusz' DOS Routines (MDR): http://mdr.osdn.io
 * Published under the terms of the MIT License, as stated below.
 *
 * Copyright (C) 2014-2023 Mateusz Viste
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/*** Excerpt from TechHelp! **************************************************
 Starting with DOS 3.0, DOS stores an additional string after the end of the
 formal environment that can be used to locate the drive and directory from
 which a program was loaded. Applications can use this to help locate program
 overlays and data files.

 This becaem important with DOS 3.0, when DOS began allowing an operator to
 invoke a program by prefixing a pathname before the program filename.

 Following the final string of the environment is a byte of 0, indicating the
 end of the formal environment. The next two bytes are a 16-bit binary count
 of additional strings (this value is normally 0001H). Following this binary
 value, you can expect to find an ASCIIZ string of the filespec (including the
 drive, path, and extension) which was used by DOS Fn 4bH (EXEC) to load and
 execute the program.

 Even if the program file was not in the default directory and DOS or Windows
 needed to use the command search path to locate the program file, the correct
 full filespec (d:\path\filename.ext,0) will be stored after the end of the
 environment.
*****************************************************************************/

#include <mdr\dos.h>

/* returns a far pointer to the full path and filename of the running program.
 * returns NULL on error. */
const char far *mdr_dos_selfexe(void) {
  char far *env;

  /* compute the env pointer */
  env = mdr_dos_env();

  /* skip all environment variables */
  for (;;) {
    while (*env != 0) env++; /* go to end of variable */
    env++;
    if (*env == 0) break; /* end of list */
  }

  /* read the WORD that indicates the amount of strings that follow - must
   * be at least 1 since the full prog name is the first one */
  env++;
  if (*((unsigned short far *)env) == 0) return(NULL);

  /* return a pointer to the the executable name */
  return(env + 2);
}
