/*
 * Mouse routines
 *
 * This file is part of the Mateusz' DOS Routines (MDR): http://mdr.osdn.io
 * Published under the terms of the MIT License, as stated below.
 *
 * Copyright (C) 2014-2023 Mateusz Viste
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#include <dos.h>

#include <mdr/mouse.h>  /* include self for control */

/* init the mouse driver (and checks for presence of mouse support at same time)
 * returns 0 if no mouse is present, and the number of buttons otherwise */
int mdr_mouse_init(void) {
  union REGS regs;
  regs.x.ax = 0;
  int86(0x33, &regs, &regs);
  if (regs.x.ax == 0xffff) return(regs.x.bx);
  return(0);
}

/* shows the mouse pointer */
void mdr_mouse_show(void) {
  union REGS regs;
  regs.x.ax = 1;
  int86(0x33, &regs, &regs);
}

/* hides the mouse pointer */
void mdr_mouse_hide(void) {
  union REGS regs;
  regs.x.ax = 2;
  int86(0x33, &regs, &regs);
}

/* get x/y coordinates of the mouse, and returns a bitfield with state of buttons */
int mdr_mouse_getstate(int *x, int *y) {
  union REGS regs;
  regs.x.ax = 3;
  int86(0x33, &regs, &regs);
  if (x) *x = regs.x.cx;
  if (y) *y = regs.x.dx;
  return(regs.x.bx);
}

/* get x/y coordinates of the mouse when the last button release occured since last check.
   returns the id of the button pressed (1 or 2), or 0 if no event occured. */
int mdr_mouse_fetchrelease(int *x, int *y) {
  int buttonid;
  union REGS regs;
  /* check left and right buttons */
  for (buttonid = 0; buttonid < 2; buttonid++) {
    regs.x.ax = 6;         /* INT 33,6 = "get mouse button release data" */
    regs.x.bx = buttonid;  /* button id (left = 0 ; right = 1) */
    int86(0x33, &regs, &regs); /* make the INT call */
    if ((regs.x.bx != 0) && ((regs.x.ax & 3) == 0)) { /* if a click occured, report it,  */
      *x = regs.x.cx;                                 /* but only if all buttons were    */
      *y = regs.x.dx;                                 /* released at that time! This is  */
      return(buttonid + 1);                           /* to avoid false clicks from some */
    }                                                 /* mouses.                         */
  }
  return(0);
}
