/*
 * Library to access OPL2/OPL3 hardware (YM3812 / YMF262)
 *
 * This file is part of the Mateusz' DOS Routines (MDR): http://mdr.osdn.io
 * Published under the terms of the MIT License, as stated below.
 *
 * Copyright (C) 2015-2023 Mateusz Viste
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#include <conio.h> /* inp(), out() */

#include <mdr/bios.h> /* mdr_bios_tickswait() */

#include <mdr/opl.h>

#define PORT_CTRL 0x388
#define PORT_DATA 0x389
#define VOICESCOUNT 9


/* tables below provide register offsets for each voice. note, that these are
 * NOT the registers IDs, but their direct offsets instead - this for simpler
 * and faster computations. */
const unsigned char op1offsets[9] = {0x00,0x01,0x02,0x08,0x09,0x0a,0x10,0x11,0x12};
const unsigned char op2offsets[9] = {0x03,0x04,0x05,0x0b,0x0c,0x0d,0x13,0x14,0x15};

/* cache for instrument's carrier_40 value, this is needed to recompute the
 * volume of a channel */
static unsigned char carr40_cache[9];


/* this is a LOW-LEVEL function that writes a data byte into the reg register
 * of the OPL chip. Use only if you know exactly what you are doing. */
void mdr_opl_regwr(unsigned char reg, unsigned char data) {
  int i;
  /* select the register we want to write to, via the index register */
  outp(PORT_CTRL, reg);
  /* OPL2 requires 3.3us to pass before writing to the data port. AdLib
   * recommends reading 6 times from the index register to make time pass. */
  i = 6;
  while (i--) inp(PORT_CTRL);
  /* write the data to the data register */
  outp(PORT_DATA, data);
  /* OPL2 requires 23us to pass before writing to the data port. AdLib
   * recommends reading 35 times from the index register to make time pass. */
  i = 35;
  while (i--) inp(PORT_CTRL);
}


/* Hardware detection and initialization.
 * Returns 0 on success, non-zero otherwise. */
int mdr_opl_init(void) {
  int x, y;

  /* detect the hardware and return error if not found */
  mdr_opl_regwr(0x04, 0x60); /* reset both timers by writing 60h to register 4 */
  mdr_opl_regwr(0x04, 0x80); /* enable interrupts by writing 80h to register 4 (must be a separate write from the 1st one) */
  x = inp(PORT_CTRL) & 0xE0; /* read the status register (port 388h) and store the result */
  mdr_opl_regwr(0x02, 0xff); /* write FFh to register 2 (Timer 1) */
  mdr_opl_regwr(0x04, 0x21); /* start timer 1 by writing 21h to register 4 */
  mdr_bios_tickswait(1); /* Creative Labs recommends a delay of at least 80 microseconds
                            DO NOT perform inp() calls for delay here, some cards do not
                            initialize well then (reported for CT2760) */
  y = inp(PORT_CTRL) & 0xE0;  /* read the upper bits of the status register */
  mdr_opl_regwr(0x04, 0x60); /* reset both timers and interrupts (see steps 1 and 2) */
  mdr_opl_regwr(0x04, 0x80); /* reset both timers and interrupts (see steps 1 and 2) */
  /* test the stored results of steps 3 and 7 by ANDing them with E0h. The result of step 3 should be */
  if (x != 0) return(-1);    /* 00h, and the result of step 7 should be C0h. If both are     */
  if (y != 0xC0) return(-2); /* ok, an AdLib-compatible board is installed in the computer   */

  mdr_opl_regwr(0x01, 0x20);  /* enable Waveform Select */
  mdr_opl_regwr(0x04, 0x00);  /* turn off timers IRQs */
  mdr_opl_regwr(0x08, 0x00);  /* turn off CSW mode and reset the octave split point */
  mdr_opl_regwr(0xBD, 0x00);  /* set vibrato/tremolo depth to low and set melodic mode */

  for (x = 0; x < VOICESCOUNT; x++) {
    mdr_opl_regwr(0x20 + op1offsets[x], 0x1);     /* set the modulator's multiple to 1 */
    mdr_opl_regwr(0x20 + op2offsets[x], 0x1);     /* set the modulator's multiple to 1 */
    mdr_opl_regwr(0x40 + op1offsets[x], 0x10);    /* set volume of all channels to about 40 dB */
    mdr_opl_regwr(0x40 + op2offsets[x], 0x10);    /* set volume of all channels to about 40 dB */
  }

  mdr_opl_clear();

  /* all done */
  return(0);
}


/* close OPL device */
void mdr_opl_close(void) {
  unsigned char x;

  /* turns all notes 'off' */
  mdr_opl_clear();

  /* set volume to lowest level on all voices */
  for (x = 0; x < VOICESCOUNT; x++) {
    mdr_opl_regwr(0x40 + op1offsets[x], 0x1f);
    mdr_opl_regwr(0x40 + op2offsets[x], 0x1f);
  }
}


/* turns off all notes */
void mdr_opl_clear(void) {
  int i;
  for (i = 0; i < VOICESCOUNT; i++) mdr_opl_noteoff(i);

  /* reset the percussion bits at the 0xBD register */
  mdr_opl_regwr(0xBD, 0);
}


/* loads an instrument described by properties in a timbre struct into
 * the defined voice channel. The OPL2 chip supports up to 9 voice channels,
 * from 0 to 8. The timbre struct can be freed right after this call. */
void mdr_opl_loadinstrument(unsigned char voice, const struct mdr_opl_timbre *timbre) {
  /* remember carrier40's KSL value for later volume computations */
  carr40_cache[voice] = timbre->car_40 & 0xC0;

  /* KSL (key level scaling) / attenuation */
  mdr_opl_regwr(0x40 + op1offsets[voice], timbre->mod_40);
  mdr_opl_regwr(0x40 + op2offsets[voice], timbre->car_40);

  /* select waveform on both operators */
  mdr_opl_regwr(0xE0 + op1offsets[voice], timbre->mod_ws);
  mdr_opl_regwr(0xE0 + op2offsets[voice], timbre->car_ws);

  /* sustain / release */
  mdr_opl_regwr(0x80 + op1offsets[voice], timbre->mod_sr);
  mdr_opl_regwr(0x80 + op2offsets[voice], timbre->car_sr);

  /* attack rate / decay */
  mdr_opl_regwr(0x60 + op1offsets[voice], timbre->mod_ad);
  mdr_opl_regwr(0x60 + op2offsets[voice], timbre->car_ad);

  /* AM / vibrato / envelope */
  mdr_opl_regwr(0x20 + op1offsets[voice], timbre->mod_20);
  mdr_opl_regwr(0x20 + op2offsets[voice], timbre->car_20);

  mdr_opl_regwr(0xC0 + voice, timbre->feedconn);
}


/* adjusts volume of a voice. volume goes from 63 (mute) to 0 (loudest) */
void mdr_opl_voicevolume(unsigned char voice, unsigned char volume) {
  mdr_opl_regwr(0x40 + op2offsets[voice], carr40_cache[voice] | volume);
}
