/*
 * IMF reader/player
 *
 * This file is part of the Mateusz' DOS Routines (MDR): http://mdr.osdn.io
 * Published under the terms of the MIT License, as stated below.
 *
 * Copyright (C) 2023 Mateusz Viste
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#include <mdr/opl.h>

/* the IMF file format is a raw stream of OPL register-value-timing tuples.
 * each tuple is 4 bytes long and consists of:
 *
 *   unsigned char   ; ADLIB REGISTER
 *   unsigned char   ; ADLIB VALUE
 *   unsigned short  ; DELAY
 *
 * The DELAY is provided as a number of "ticks", while the length of a tick
 * is undefined and must be known in advance by the IMF-playing implementation.
 * Commonly found values in games are 560 and 700 Hz.
 */


struct IMFCHUNK {
  unsigned char adreg;
  unsigned char adval;
  unsigned short delay;
};

static unsigned short next_wakeup = 0;


/* playback initialization, this function must be called immediately before
 * playback. imf points to the start of the IMF file and must contain at least
 * the first 6 bytes of the audio file.
 * clock must be an incrementing value that wraps to 0 after 65535.
 * the clock speed will control the playback's tempo.
 * returns the amount of consumed bytes (0, 4 or 6) */
unsigned short mdr_opl_imf_init(void *imf, unsigned short clock) {
  next_wakeup = clock;
  if (((short *)imf)[0] == 0) { /* type-0 IMF (no header, only the 4-bytes zeroed 1st instruction) */
    return(4);
  } else if (((short *)imf)[1] == 0) { /* type-1 (2-bytes header) */
    return(6);
  } else {
    return(0);
  }
}


/* Playback, advanced version. Feeds the IMF playback routine with IMF data.
 * Returns the amount of bytes that have been consumed, hence the next call
 * should provide an imf pointer advanced by this many bytes (and imflen
 * decreased accordingly). Such approach might not be the most intuitive, but
 * it allows to load an imf song partially and provide only short chunks of
 * data for playback instead of having to buffer the entire song in memory.
 * For a simpler call that requires to buffer the entire IMF file in memory,
 * see mdr_opl_imf_playeasy().
 * This function must be called repeatedly at a high frequency for best
 * playback quality. */
unsigned short mdr_opl_imf_play(void *imf, unsigned short imflen, unsigned short clock) {
  unsigned short consumedbytes = 0;
  struct IMFCHUNK *imfnode = imf;

  while (imflen >= 4) {

    /* it's not time yet (unless timer wraped) */
    if (clock < next_wakeup) {
      /* detect time wrap */
      if (next_wakeup - clock < 16384) return(consumedbytes);
    } else if (clock - next_wakeup > 16384) {
      return(consumedbytes);
    }

    /* ignore writes to reg 0: such register does not exist on OPL2, it is
     * likely some kind of marker in the IMF file */
    if (imfnode->adreg != 0) mdr_opl_regwr(imfnode->adreg, imfnode->adval);

    next_wakeup += imfnode->delay;

    imfnode++;
    consumedbytes += 4;
    imflen -= 4;
  }

  return(consumedbytes);
}
