/*
 * IMF reader/player (easy interface, but requires to buffer the entire IMF)
 *
 * This file is part of the Mateusz' DOS Routines (MDR): http://mdr.osdn.io
 * Published under the terms of the MIT License, as stated below.
 *
 * Copyright (C) 2023 Mateusz Viste
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#include <mdr/opl.h>

static unsigned char *imfptrorg;
static unsigned short imflenorg;
static unsigned char *imfptr;
static unsigned short imflen;
static unsigned char loops;

/* playback initialization, easy interface. imf points to the start of the IMF
 * file. The imf pointer must not be freed as long as playback is ongoing.
 * imflength is the size (in bytes) of the IMF data.
 * clock must be an incrementing value that wraps to 0 after 65535. The clock
 * speed will control the playback's tempo.
 * loopscount tells how many times the song will have to be looped (0 means
 * "loop forever").
 * returns 0 on success, non-zero otherwise. */
int mdr_opl_imfeasy_init(void *imf, unsigned short imflength, unsigned short clock, unsigned char loopscount) {
  unsigned short bytes;
  if ((imflength < 8) || (imf == (void *)0)) return(-1);

  imflenorg = imflength;
  loops = loopscount;

  bytes = mdr_opl_imf_init(imf, clock);
  imfptrorg = ((unsigned char *)imf) + bytes;
  imflenorg = imflength - bytes;
  imfptr = imfptrorg;
  imflen = imflenorg;
  return(0);
}


/* Playback of an IMF file preloaded via mdr_opl_imfeasy_init(). This function
 * must be called repeatedly at a high frequency for best playback quality.
 * Returns 0 on success, 1 if playback ended, -1 on error. */
int mdr_opl_imfeasy_play(unsigned short clock) {
  unsigned short bytes;

  if (imflen < 3) {
    if (imflenorg == 0) return(-1);
    if (loops == 1) return(1);
    if (loops > 0) loops--;
    imfptr = imfptrorg;
    imflen = imflenorg;
  }

  bytes = mdr_opl_imf_play(imfptr, imflen, clock);
  imflen -= bytes;
  imfptr += bytes;

  return(0);
}
