/*
 * OPL2 notes driving (note ON / note OFF / pitch bending)
 *
 * This file is part of the Mateusz' DOS Routines (MDR): http://mdr.osdn.io
 * Published under the terms of the MIT License, as stated below.
 *
 * Copyright (C) 2015-2023 Mateusz Viste
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#include <mdr/opl.h>

/* cache for currently playing notes, required for computing KEY-OFF events */
static unsigned char keyon_cache[9];

/* releases a note on selected voice. */
void mdr_opl_noteoff(unsigned char voice) {
  mdr_opl_regwr(0xB0 + voice, keyon_cache[voice]);
}


/* Triggers a note on selected voice channel.
 * freqid is a value between 0 and 1023. The following formula can be used to
 * determine the freq number for a given note frequency (Hz) and block:
 *
 *   freqid = frequency * 2^(20 - block) / 49716
 *
 * The note will be kept "pressed" until mdr_opl_noteoff() is called. */
void mdr_opl_noteon(unsigned char voice, unsigned short freq, enum mdr_opl_fgroup_t fgroup) {
  keyon_cache[voice] = fgroup | (freq >> 8);
  mdr_opl_regwr(0xA0 + voice, freq & 0xff); /* set lowfreq */
  mdr_opl_regwr(0xB0 + voice, keyon_cache[voice] | 32); /* KEY ON + hifreq + octave */
}


/* changes the frequency of the note currently playing on voice channel, this
 * can be used for pitch bending. */
void mdr_opl_notebend(unsigned char voice, unsigned short freq, enum mdr_opl_fgroup_t fgroup) {
  keyon_cache[voice] = fgroup | (freq >> 8);
  mdr_opl_regwr(0xA0 + voice, freq & 0xff); /* set lowfreq */
  mdr_opl_regwr(0xB0 + voice, (freq >> 8) | fgroup); /* hifreq + octave */
}
