/*
 * OPL2/OPL3 Timers
 *
 * This file is part of the Mateusz' DOS Routines (MDR): http://mdr.osdn.io
 * Published under the terms of the MIT License, as stated below.
 *
 * Copyright (C) 2015-2023 Mateusz Viste
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */


/* the OPL2 hardware features two high-resolution timers:
 *
 * timer 1: 12.5 kHz
 * this is an 8-bit clock that counts forward at a rate of 12.5 kHz (ie. a
 * single step duration is 80us). It takes the value preset in register 02,
 * increments it until it overflows and sets the 7th bit of the status register
 * (flag 0x40 in reg 1).
 *
 * timer 2: 3.125 kHz
 * Same as timer 1, but with a different frequency (each step duration is
 * 320 us), preset register is register 03 and overflow sets the 6th bit of the
 * status register (flag 0x20 in reg 1).
 *
 * Resetting the "timer overflow" flag in the status register is done by
 * writing 0x80 to register 04. This resets both Timer 1 and Timer 2.
 *
 * To initialize (start) Timer 1, write 0x01 to register 04.
 * To initialize (start) Timer 2, write 0x02 to register 04.
 *
 * Because the timer overflow reset flag is shared by the two timers, it is
 * highly impractical to use both timers at the same time.
 */

#include <conio.h> /* inp() */

#include <mdr/opl.h>


/* configures and starts a timer given type so it emits a tick every count
 * periods. Two timer types are available:
 *   MDR_OPL_TIMER_80US  - with a period of 80us
 *   MDR_OPL_TIMER_320US - with a period of 320us
 * count may range from 0 to 255, but 0 means "256 periods".
 *
 * You may use only one timer at a time.
 *
 * EXAMPLE: setting up MDR_OPL_TIMER_80US with a count of 25 would make the
 *          timer tick every 2ms (25 * 80us). */
void mdr_opl_timer_set(enum MDR_OPL_TIMER timertype, unsigned char count) {
  /* stop the timers */
  mdr_opl_regwr(4, 0);

  /* reset timer's flags */
  mdr_opl_regwr(4, 0x80);

  if (timertype == MDR_OPL_TIMER_80US) {
    mdr_opl_regwr(2, 256 - count); /* preset val into reg 02 (timer1) */
    mdr_opl_regwr(4, 0x01);        /* start the timer */
  } else {
    mdr_opl_regwr(3, 256 - count); /* preset val into reg 03 (timer2) */
    mdr_opl_regwr(4, 0x02);        /* start the timer */
  }
}


/* returns 1 if timer tick occured, 0 otherwise. After a tick has been
 * returned, this function will return 0 until next tick.
 *
 * it is important to note that there is no way to know whether one tick
 * passed since last time, or more, so it is up to you to make sure you call
 * this function fast enough. */
unsigned char mdr_opl_timer_tick(void) {
  /* read status register, is T1 or T2 flag ON? No = abort early */
  if ((inp(0x388) & 0x60) == 0) return(0);

  /* flag set = reset the timer and return 1 */
  mdr_opl_regwr(4, 0x80);
  return(1);
}
