/*
 * PCX-loading routines
 *
 * This file is part of the Mateusz' DOS Routines (MDR): http://mdr.osdn.io
 * Published under the terms of the MIT License, as stated below.
 *
 * Copyright (C) 2022-2023 Mateusz Viste
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#include <stdio.h>
#include <string.h>

#include <mdr/pcx.h>


/* analyzes the header of a PCX file and fills the pcx_hdr struct accordingly.
 * fd      must be a valid (open) file descriptor.
 * offset  is the address inside the file where the PCX data is located
 *         (usually 0, unless the file is some kind of container).
 * len     is the total length of the PCX data. len=0 means "same as file size"
 * returns 0 on success, non-zero otherwise. */
int mdr_pcx_anal(struct pcx_hdr *h, FILE *fd, unsigned long offset, unsigned long len) {
  struct {
    unsigned char sig;        /* always 0x0A */
    unsigned char version;    /* 0-5 */
    unsigned char rleflag;    /* should always be set to 1 (RLE) */
    unsigned char bitsperplane; /* 1, 2, 4 or 8 */
    unsigned short min_x;
    unsigned short min_y;
    unsigned short max_x;
    unsigned short max_y;
    unsigned short dpi_h;     /* horizontal resolution (DPI) */
    unsigned short dpi_v;     /* vertical resolution (DPI) */
    unsigned char egapal[48]; /* EGA palette */
    unsigned char reserved1;
    unsigned char planes_per_pix; /* number of color planes per pixel (1, 3 or 4) */
    unsigned short bytes_per_scanline; /* number of bytes *of a single plane* per scanline */
    unsigned short paltype;   /* palette type, 1=mono or color ; 2=grayscale */
  } hdr;

  /* clear out the pcx_hdr struct */
  bzero(h, sizeof(*h));

  /* read hdr */
  fseek(fd, offset, SEEK_SET);
  if (fread(&hdr, 1, sizeof(hdr), fd) != sizeof(hdr)) return(-2);

  /* validate the file format is good and within the range of supported values */
  if (hdr.sig != 0x0A) return(-100);
  if (hdr.version != 5) return(-101);
  if (hdr.rleflag > 1) return(-102);
  if (hdr.planes_per_pix != 1) return(-103);
  if ((hdr.bitsperplane != 8) && (hdr.bitsperplane != 4) && (hdr.bitsperplane != 2) && (hdr.bitsperplane != 1)) return(-104);

  /* normalize min/max x/y so it always starts at 0 */
  hdr.max_x -= hdr.min_x;
  hdr.max_y -= hdr.min_y;
  hdr.min_x = 0;
  hdr.min_y = 0;

  /* if 8bpp then jump to palette (first byte should be 0x0C, followed by 256 RGB triplets) */
  if (hdr.bitsperplane == 8) {
    unsigned short i;
    if (len != 0) {
      fseek(fd, offset + len - 769, SEEK_SET);
    } else { /* len=0 means "length is same as file size" */
      fseek(fd, -769, SEEK_END);
    }
    if (getc(fd) != 0x0C) {
      fclose(fd);
      return(-105);
    }

    /* read the palette entries one by one and apply them to VGA, note that
     * entries are offset by 1 to make room for a custom #0 color */
    for (i = 0; i < 256; i++) {
      h->pal[i].r = getc(fd);
      h->pal[i].g = getc(fd);
      h->pal[i].b = getc(fd);
    }
  } else { /* 1, 2 and 4bpp palettes are stored in the PCX header */
    memcpy(h->pal, hdr.egapal, 48);
  }

  /* fill in the pcx_hdr struct */
  h->rle = hdr.rleflag;
  h->bpp = hdr.bitsperplane;
  h->max_x = hdr.max_x;
  h->max_y = hdr.max_y;
  h->bytes_per_scanline = hdr.bytes_per_scanline;

  /* jump to start of image data so we're ready for pcx_loadrow_next() calls */
  if (fseek(fd, offset + 128, SEEK_SET) != 0) return(-1);

  return(0);
}


/* this function should be called to load the next row of a PCX file into a
 * buffer pointed at by bufptr. you will typically want to call this function
 * h->max_y times. ptr must be at least (h->max_x + 1) bytes large for 8bpp.
 * the pcx_hdr struct must have been produced by pcx_anal().
 * returns 0 on success, non-zero otherwise. */
int mdr_pcx_loadrow(void *bufptr, const struct pcx_hdr *h, FILE *fd) {
  unsigned char *outptr = bufptr;
  unsigned short bytes_decoded = 0;

  while (bytes_decoded < h->bytes_per_scanline) { /* decode a single scanline */
    int c;
    unsigned char rle;
    c = getc(fd);
    if (c == EOF) return(-2); /* early EOF (file corrupted, likely truncated) */
    if ((c >= 192) && (h->rle == 1)) { /* RLE value */
      rle = c & 63;
      c = getc(fd);
    } else { /* pixel data */
      rle = 1;
    }

    /* */
    for (; rle > 0; rle--) {
      if (bytes_decoded <= h->max_x) {
        *outptr = c;
        outptr++;
      }
      bytes_decoded++;
    }
  }

  return(0);
}
