/*
 * Reading from and writing to an RS-232 port
 *
 * This file is part of the Mateusz' DOS Routines (MDR): http://mdr.osdn.io
 * Published under the terms of the MIT License, as stated below.
 *
 * Copyright (C) 2015-2022 Mateusz Viste
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#include <conio.h>
#include <dos.h>

#include <mdr\rs232.h>

/* get the I/O port for COMx (1..4) */
unsigned short rs232_getport(int x) {
  unsigned short far *bioscomtable = MK_FP(0,0x400);
  unsigned short res;
  if (x < 1) return(0);
  if (x > 4) return(0);
  res = bioscomtable[x - 1];
  return(res);
}

/* check if the COM port is ready for write. loops for some time waiting.
 * returns 0 if port seems ready eventually, non-zero otherwise. can be used
 * to verify the rs232 presence */
int rs232_check(unsigned short port) {
  int i = 4096; /* cycles up to 4K times (should be enough for any UART) */
  while (i-- > 0) {
    if ((inp(port + 5) & 0x20) != 0) return(0);
  }
  /* the port wouldn't become ready - exit with error */
  return(-1);
}

/* write a byte to the COM port at 'port'. this function will block if the
 * UART is not ready to transmit yet. */
void rs232_write(unsigned short port, int data) {
  /* wait for the UART to become ready for write */
  while ((inp(port + 5) & 0x20) == 0);
  /* write the data byte now */
  outp(port, data);
}

/* read a byte from COM port at 'port'. returns the read byte, or -1 if
 * nothing was available to read. */
int rs232_read(unsigned short port) {
  /* if nothing awaits, return -1 */
  if ((inp(port + 5) & 0x01) == 0) return(-1);
  /* otherwise read from port and return the result */
  return(inp(port));
}
