/*
 * SoundBlaster routines for DSP driving
 *
 * Copyright (C) 2022 Mateusz Viste
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#include <conio.h>  /* inp() and outp() */
#include <stdlib.h> /* NULL, free(), ... */

#include <mdr\sbdigi.h>


struct sbdigi_ctx {
  unsigned short port;
  unsigned char irq;
  unsigned char dma;
};


/* convenience defines for the I/O addresses of the SoundBlaster */
#define IO_RESET  (ctx->port+0x06)
#define IO_READ   (ctx->port+0x0A)
#define IO_WRITE  (ctx->port+0x0C)
#define IO_WRSTAT (ctx->port+0x0C) /* same as IO_WRITE, yes, but for inputs only */
#define IO_RDSTAT (ctx->port+0x0E)


/* address of the previous int handler */
//static void (__interrupt __far *prev_inthandler)();


/* wait for at one clock tick to pass (ie. between 0 and 55ms) */
static void waitatick(void) {
  _asm {
    xor ah, ah
    int 0x1A  /* number of clock ticks since midnight is returned in CX:DX */
    mov bx, dx /* save ticks value in BX and poll again until value change */
    POLL:
    int 0x1A
    cmp dx, bx
    je POLL
  }
}


/* read the soundblaster configuration from the BLASTER variable, returns 0 on success
 * Example of the BLASTER variable:
 * A220 I7 D1 T6 P330 H5
 *  |   |  |  |   |   +--> High (16-bit) DMA channel
 *  |   |  |  |   +------> MPU Port
 *  |   |  |  +----------> Type
 *  |   |  +-------------> DMA channel
 *  |   +----------------> IRQ
 *  +--------------------> Port
 */
static struct sbdigi_ctx *parseblasterconf(const char *blaster) {
  long A = 0, I = 0, D = 0;
  struct sbdigi_ctx *ctx;

  if (blaster == NULL) return(NULL);
  /* look for the "A/I/D" letters in the env variable */
  for (; *blaster != 0; blaster++) {
    if (*blaster == 'A') A = strtol(blaster + 1, NULL, 16); /* port addr is in hex */
    if (*blaster == 'I') I = strtol(blaster + 1, NULL, 10);
    if (*blaster == 'D') D = strtol(blaster + 1, NULL, 10);
  }
  /* check values */
  if ((A < 1) || (I < 1) || (I > 7) || (D < 1) || (D > 3)) return(NULL);

  ctx = malloc(sizeof(struct sbdigi_ctx));
  if (ctx == NULL) return(NULL);

  ctx->port = (unsigned short)A;
  ctx->irq = (unsigned char)I;
  ctx->dma = (unsigned char)D;

  return(ctx);
}


static int sbdigi_dsp_read(struct sbdigi_ctx *ctx) {
  /* When DSP data is available, it can be read in from the Read Data port.
   * Before the data is read in, bit-7 of the Read-Buffer Status port must be
   * checked to ensure that there is data to read. If bit-7 is 1, then there
   * is data to read. Otherwise, no data is available. */
  if ((inp(IO_RDSTAT) & 128) == 0) return(-1); /* nothing to read */
  return(inp(IO_READ));
}


static void sbdigi_dsp_write(struct sbdigi_ctx *ctx, int databyte) {
  /* DSP commands and data are sent through the Write Command/Data port.
   * Before data is written to the DSP, bit-7 of the Write-Buffer Status port
   * must be checked to ensure that the DSP command/data buffer is empty. If
   * bit-7 is 0, the DSP buffer is empty and is ready to receive commands or
   * data. Otherwise, no commands or data should be written to the DSP. */
  while ((inp(IO_WRSTAT) & 128) != 0); /* wait for the 'write' buffer to become available */
  outp(IO_WRITE, databyte);
}


/* ISR for SB IRQ handling
 * - tell the Sound Blaster DSP to stop pulling the interrupt line by
 *   performing a single read on the DATA_AVAILABLE port (2xEh)
 * - send an EOI to the PIC
 * - return from interrupt (IRET instruction) */
//void __interrupt __far sbdigi_irqhandler(union INTPACK r) {
  /* perform a single read of the dsp */
//  inp(IO_RDSTAT); /* I need to know the port here... */

  /* EOI */
//  outp(0x20, 0x20);

  /* this function ends with an IRET (done by OpenWatcom because of the
   * __interrupt flag) */
//}


/* BLASTER lookup, DSP reset and speakers ON */
struct sbdigi_ctx *sbdigi_init(const char *blaster) {
  /* 1. Write a "1" to the Reset port (2x6h) and wait for 3 microseconds.
   * 2. Write a "0" to the Reset port.
   * 3. Poll for a ready byte 0AAh from the Read Data port. You must check the
   *    the Read-Buffer Status port to ensure there is data before reading the
   *    Read Data port. */
  int x, timeout;
  struct sbdigi_ctx *ctx;

  /* look for the BLASTER environment variable */
  ctx = parseblasterconf(blaster);
  if (ctx == NULL) return(NULL);

  outp(IO_RESET, 1);
  waitatick(); /* must wait at least 3 microseconds */
  waitatick();
  outp(IO_RESET, 0);
  /* wait for the DSP to give us the green light - no longer than 50ms (the
   * SoundBlaster docs state that the DSP should init in about 100us) */
  for (timeout = 4; timeout != 0; timeout--) {
    waitatick(); /* don't busy poll the DSP as this blocks some sb clones */
    /* check that we got something on the data port */
    x = sbdigi_dsp_read(ctx);
    if (x < 0) continue; /* nothing available yet */
    if (x == 0xAA) goto installisr; /* 0xAA means 'OK' */
    /* otherwise there's a problem */
    break;
  }
  /* if we are here, then a timeout occured */
  free(ctx);
  return(NULL);

  /* sb initialized fine, install the ISR now */
  installisr:
/*  prev_inthandler = _dos_getvect(irq);
  _dos_setvect(irq, sbdigi_irqhandler);*/

  sbdigi_spkon(ctx);
  return(ctx);
}


void sbdigi_playsample(struct sbdigi_ctx *ctx, void *ptr, unsigned short len) {
  /* some of the ports differ depending on the target DMA channel:
   * DMA  PAGE  ADDR  LENGTH
   *  0   0x87   0     1
   *  1   0x83   2     3
   *  2   0x81   4     5
   *  3   0x82   6     7
   */
  unsigned long ptr_phys;
  unsigned char ptr_page, ptr_pos_lo, ptr_pos_hi;
  const unsigned char dreg_page[] = {0x87, 0x83, 0x81, 0x82};

  /* convert pointer into its physical address and compute its page and
   * high/low bytes of the position offset */
  ptr_phys = (((unsigned long)ptr >> 16) * 0x10) + ((unsigned long)ptr & 0xffff);
  ptr_page = ptr_phys >> 16;
  ptr_pos_lo = ptr_phys & 0xff;
  ptr_pos_hi = (ptr_phys >> 8) & 0xff;

  /* prep the dma channel */
  outp(0x0A, 4 | ctx->dma);    /* disable channel (chan number | 4) */
  outp(0x0C, 0);               /* trigger the flip flop (write anything there) */
  outp(0x0B, 0x48 | ctx->dma); /* transfer mode=SINGLE (0x48 + channel number) */
  outp(dreg_page[ctx->dma], ptr_page);  /* page transfer */
  outp(ctx->dma * 2, ptr_pos_lo);       /* position low bit */
  outp(ctx->dma * 2, ptr_pos_hi);       /* position high bit */
  outp((ctx->dma * 2) + 1, len & 0xff); /* count low byte */
  outp((ctx->dma * 2) + 1, len >> 8);   /* count high byte */
  outp(0x0A, ctx->dma);        /* enable requested channel */

  /* prep the SB DSP */
  len--; /* the SB DSP expects length to be one less (so 0xffff means 65536 bytes) */
  sbdigi_dsp_write(ctx, 0x40); /* set time constant: (65536 - (256000000 / (channels * sample_rate))) >> 8 */
  sbdigi_dsp_write(ctx, 0x83); /* 8000 hz */
  sbdigi_dsp_write(ctx, 0x14); /* I/O command for 8-bit PCM output */
  sbdigi_dsp_write(ctx, len & 0xff); /* followed by the length-1 (low byte, then high byte) */
  sbdigi_dsp_write(ctx, len >> 8);
  /* transfer begins here, DSP will generate an interrupt once transfer is complete */
}


void sbdigi_spkon(struct sbdigi_ctx *ctx) {
  sbdigi_dsp_write(ctx, 0xD1);
}


void sbdigi_spkoff(struct sbdigi_ctx *ctx) {
  sbdigi_dsp_write(ctx, 0xD3);
}


void sbdigi_quit(struct sbdigi_ctx *ctx) {
  sbdigi_spkoff(ctx);
  free(ctx);
}
