/*
 * Routines for computation of basic transcendental functions sin and cos.
 * These routines use only integers, hence they do not require an FPU nor any
 * kind of FPU emulation. Works reasonably fast even on an 8086 CPU.
 *
 * The results are computed using polynomial approximations. Their precision
 * is not expected to be ideal, but "good enough" for common usage.
 *
 * This file is part of the Mateusz' DOS Routines (MDR): http://mdr.osdn.io
 * Published under the terms of the MIT License, as stated below.
 *
 * Copyright (C) 2022 Mateusz Viste
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#include <mdr\trigint.h>

/* helper definitions for common "angle to rad*1000" values */
#define ANGLE_360 6283
#define ANGLE_315 5498
#define ANGLE_270 4712
#define ANGLE_225 3927
#define ANGLE_180 3142
#define ANGLE_135 2357
#define ANGLE_90  1571
#define ANGLE_45   785


short trigint_sin(short rad1000) {
  unsigned long res;
  short sign = 1; /* signum of the result (1 or -1) */

  /* if angle is over 360 deg (6.283 rads), bring it back in range 0..360 */
  rad1000 %= ANGLE_360;
  if (rad1000 < 0) rad1000 += ANGLE_360; /* normalize negative values */

  if (rad1000 <= ANGLE_45) { /* 0..45 deg */
    /* nothing to prep, jump to computation right away */
  } else if (rad1000 <= ANGLE_90) { /* 45..90 deg -> cos(90-x) */
    return(trigint_cos(ANGLE_90 - rad1000));
  } else if (rad1000 <= ANGLE_135) { /* 90..135 deg -> sin(180-x) -> cos(90-(180-x)) */
    return(trigint_cos(ANGLE_90 - (ANGLE_180 - rad1000)));
  } else if (rad1000 <= ANGLE_180) { /* 135..180 deg -> sin(180-x) */
    rad1000 = ANGLE_180 - rad1000;
  } else if (rad1000 <= ANGLE_225) { /* 180..225 deg -> -sin(x-180) */
    sign = -1;
    rad1000 -= ANGLE_180;
  } else if (rad1000 <= ANGLE_270) { /* 225..270 deg -> -sin(x-180) = -cos(90-(x-180)) */
    return(-trigint_cos(ANGLE_90 - (rad1000 - ANGLE_180)));
  } else if (rad1000 <= ANGLE_315) { /* 270..315 deg -> sin(360-x) = -cos(90-(360-x)) */
    return(-trigint_cos(ANGLE_90 - (ANGLE_360 - rad1000)));
  } else { /* 315..360 deg -> -sin(360-x) */
    sign = -1;
    rad1000 = ANGLE_360 - rad1000;
  }

  /* leverage a simplified polynomial approximation of the sine (works well
   * only for the range 0..PI/4, hence the above normalization steps):
     sin(x) = x - (x^3 / 6) */
  res = rad1000;
  res *= rad1000;
  res *= rad1000;
  res /= 6000000l;

  return(sign * (rad1000 - res));
}


short trigint_cos(short rad1000) {
  short sign = 1; /* signum of the result (1 or -1) */
  unsigned long resa, resb;

  /* if angle is over 360 deg (6.283 rads), bring it back in range 0..360 */
  rad1000 %= ANGLE_360;
  if (rad1000 < 0) rad1000 += ANGLE_360; /* normalize negative values */

  if (rad1000 <= ANGLE_45) { /* 0..45 deg */
    /* nothing to prep, jump to computation right away */
  } else if (rad1000 <= ANGLE_90) { /* 45..90 deg -> sin(90-x) */
    return(trigint_sin(ANGLE_90 - rad1000));
  } else if (rad1000 <= ANGLE_135) { /* 90..135 deg -> -cos(180-x) -> -sin(90-(180-x)) */
    return(-trigint_sin(ANGLE_90 - (ANGLE_180 - rad1000)));
  } else if (rad1000 <= ANGLE_180) { /* 135..180 deg -> -cos(180-x) */
    sign = -1;
    rad1000 = ANGLE_180 - rad1000;
  } else if (rad1000 <= ANGLE_225) { /* 180..225 deg -> -cos(x-180) */
    sign = -1;
    rad1000 -= ANGLE_180;
  } else if (rad1000 <= ANGLE_270) { /* 225..270 deg -> -cos(x-180) = -sin(90-(x-180)) */
    return(-trigint_sin(ANGLE_90 - (rad1000 - ANGLE_180)));
  } else if (rad1000 <= ANGLE_315) { /* 270..315 deg -> cos(360-x) = sin(90-(360-x)) */
    return(trigint_sin(ANGLE_90 - (ANGLE_360 - rad1000)));
  } else { /* 315..360 deg -> cos(360-x) */
    rad1000 = ANGLE_360 - rad1000;
  }

  /* leverage a simplified polynomial approximation of the cosine (works well
   * only for the range 0..PI/4, hence the above normalization steps):
   * cos(x) = 1 - (x^2 / 2) + (x^4 / 24) */
  resa = rad1000;
  resa *= rad1000;
  resa /= 2000;

  resb = rad1000;
  resb *= rad1000;
  resb *= rad1000;
  resb /= 1000;
  resb *= rad1000;
  resb /= 24000000l;

  return(sign * (1000 + resb - resa));
}
