/*
 * Function to iterate through files in a ZIP archive.
 *
 * This file is part of the Mateusz' DOS Routines (MDR): http://mdr.osdn.io
 * Published under the terms of the MIT License, as stated below.
 *
 * Copyright (C) 2012-2022 Mateusz Viste
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#include <stdio.h>     /* FILE */
#include <stdlib.h>    /* NULL */
#include <string.h>    /* memset() */

#include <mdr\unzip.h>  /* include self for control */

/* returns next item found in zip file. this is supposed to be called
 * iteratively, passing the previous mdr_zipitem struct each time (z must be
 * all zeroed out on first call).
 * returns 0 on success, neg on error, 1 on end of archive */
int mdr_zip_iter(struct mdr_zip_item *z, FILE *fd) {
  unsigned char hdrbuff[32];
  unsigned long entrysig;

  if ((fd == NULL) || (z == NULL)) return(-1);

  REPEAT:

  /* set file pointer */
  if (fseek(fd, z->nextidxoffset, SEEK_SET) != 0) return(-1);

  /* zero out the struct */
  memset(z, 0, sizeof(*z));

  /* read entry's 32-bit signature (big endian) */
  if (fread(hdrbuff, 1, 4, fd) != 4) return(-1);
  entrysig = hdrbuff[3];
  entrysig <<= 8;
  entrysig |= hdrbuff[2];
  entrysig <<= 8;
  entrysig |= hdrbuff[1];
  entrysig <<= 8;
  entrysig |= hdrbuff[0];

  /* optional data descriptor header: no need to read it, just skip it */
  if (entrysig == 0x08074b50ul) {
    z->nextidxoffset += 12; /* header is 3x4 bytes (CRC + compressed len + uncompressed len) */
    goto REPEAT;
  }

  /* central directory: I don't care about it, just consider it a proper ending of the ZIP archive */
  if (entrysig == 0x02014b50ul) return(1);

  /* local file */
  if (entrysig == 0x04034b50ul) {
    unsigned short generalpurposeflags;
    unsigned short len;
    unsigned char i;

    /* read and parse the zip header */
    if (fread(hdrbuff, 1, 26, fd) != 26) return(-2);

    /* read filename's length and copy it over */
    len = (hdrbuff[23] << 8) | hdrbuff[22];
    if (len > sizeof(z->fname) - 1) return(-3);
    if (fread(z->fname, 1, len, fd) != len) return(-4);

    /* poor man's dir detection: if fname ends with a / then it's a dir (the
     * ZIP spec forbids the usage of '\' in ZIP paths) */
    if (z->fname[len - 1] == '/') z->flags |= ZIP_FLAG_ISADIR;

    /* general-purpose flags */
    generalpurposeflags = hdrbuff[2] | (hdrbuff[3] << 8);
    if (generalpurposeflags & 1) z->flags |= ZIP_FLAG_ENCRYPTED;

    /* compression method */
    z->compmethod = hdrbuff[4] | (hdrbuff[5] << 8);

    /* last modif. date and time (DOS format, 16-bit packed) */
    z->dostime = hdrbuff[6] | (hdrbuff[7] << 8);
    z->dosdate = hdrbuff[8] | (hdrbuff[9] << 8);

    /* CRC32 */
    for (i = 13; i >= 10; i--) {
      z->crc32 <<= 8;
      z->crc32 |= hdrbuff[i];
    }

    /* compressed file's length */
    for (i = 17; i >= 14; i--) {
      z->compressedfilelen <<= 8;
      z->compressedfilelen |= hdrbuff[i];
    }

    /* true (uncompressed) file's length */
    for (i = 21; i >= 18; i--) {
      z->filelen <<= 8;
      z->filelen |= hdrbuff[i];
    }

    /* printf("Filename len: %d / extrafield len: %d / compfile len: %ld / filelen: %ld\n", len, extrafieldlen, z->compressedfilelen, z->filelen); */

    /* printf("Filename: %s (%ld bytes compressed)\n", z->filename, z->compressedfilelen); */

    len = hdrbuff[24] | (hdrbuff[25] << 8);

    z->dataoffset = ftell(fd) + len;
    z->nextidxoffset = z->dataoffset + z->compressedfilelen;

    return(0);
  }

  /* anything else is either an unsupported block or a file corruption */
  return(-5);
}
