/*
 * video library - provides a few functions for mode 4 and 13h programming.
 *
 * This file is part of the Mateusz' DOS Routines (MDR): http://mdr.osdn.io
 * Published under the terms of the MIT License, as stated below.
 *
 * Copyright (C) 2014-2022 Mateusz Viste
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#include <dos.h>       /* provides int86() along with the union REGS type */
#include <malloc.h>    /* _fcalloc() and friends */
#include <stdlib.h>
#include <string.h>    /* memcpy() */

#include <mdr\video.h> /* include self for control */


/* init video mode */
struct video_handler *video_open(int mode, int flags) {
  union REGS regs;
  struct video_handler *result;
  /* validate the mode */
  if ((mode != 0x04) && (mode != 0x13)) return(NULL);
  result = malloc(sizeof(struct video_handler));
  if (result == NULL) return(NULL);
  result->flags = flags;
  result->mode = mode;
  if (mode == 0x13) {  /* VGA mode 0x13 */
      if (flags & VIDEO_DBUF) { /* VGA with double buffering */
          result->dbuf = _fcalloc(64000u, 1);
          if (result->dbuf == NULL) {
            free(result);
            return(NULL);
          }
        } else { /* VGA without double buffering */
          result->dbuf = (unsigned char far*)0xA0000000l;
      }
    } else if (mode == 0x04) {  /* CGA mode 0x04 */
      if (flags & VIDEO_DBUF) { /* CGA with double buffering */
          result->dbuf = _fcalloc(16192, 1);
          if (result->dbuf == NULL) {
            free(result);
            return(NULL);
          }
        } else { /* CGA without double buffering */
          result->dbuf = (unsigned char far*)0xB8000000l;
      }
  }
  /* first save the current mode, to be able to restore it later */
  regs.h.ah = 0x0F;
  int86(0x10, &regs, &regs);
  result->lastmode = regs.h.al;
  /* now switch to the selected video mode */
  regs.h.ah = 0;
  regs.h.al = mode;
  int86(0x10, &regs, &regs);
  return(result);
}


/* Wait until VBLANK - works both in mode 0x04 and mode 0x13 */
void video_waitvblank(void) {
  _asm {
    mov dx, 0x3DA
    /* wait for "no vblank" (in case there is a vblank right now) */
    WAIT_NOVBLANK:
    in al, dx
    and al, 8
    jnz WAIT_NOVBLANK
    /* wait for "vblank" */
    WAIT_VBLANK:
    in al, dx
    and al, 8
    jz WAIT_VBLANK
  }
}

/* copies line ysrc over ydst in CGA mode memory */
void video_cgalinecopy(struct video_handler *handler, int ydst, int ysrc) {
  unsigned int offsetsrc = ((ysrc & 1) << 13) + ((ysrc >> 1) << 6) + ((ysrc >> 1) << 4);
  unsigned int offsetdst = ((ydst & 1) << 13) + ((ydst >> 1) << 6) + ((ydst >> 1) << 4);
  _fmemcpy(handler->dbuf + offsetdst, handler->dbuf + offsetsrc, 80);
}

/* clear screen using color */
void video_cls(struct video_handler *handler, unsigned char color) {
  if (handler->mode == 0x13) { /* VGA mode 0x13 */
      _fmemset(handler->dbuf, color, 64000u);
    } else if (handler->mode == 0x04) { /* CGA mode 0x04 */
      int colbyte;
      color &= 3;
      colbyte = color | (color << 2) | (color << 4) | (color << 6);
      _fmemset(handler->dbuf, colbyte, 8000);
      _fmemset(handler->dbuf + 8192, colbyte, 8000);
  }
}

void video_close(struct video_handler *handler) {
  union REGS regs;
  /* restore the initial video mode */
  regs.h.ah = 0;
  regs.h.al = handler->lastmode;
  int86(0x10, &regs, &regs);
  /* free the double buffering memory, if any */
  if (handler->flags & VIDEO_DBUF) _ffree(handler->dbuf);
  free(handler);
}

static void video_rputpixel_cga(struct video_handler *handler, int x, int y, unsigned char col, int repeat) {
  static unsigned char mask[4] = {63, 207, 243, 252}; /* using static here */
  static int bitshl[4] = {6, 4, 2, 0};                /* improves speed by 20% */
  unsigned int offset = ((y & 1) << 13) + ((y >> 1) << 6) + ((y >> 1) << 4);
  col &= 3;
  while (repeat-- > 0) {
    y = offset + (x >> 2); /* I reuse the y register because I won't use its value anymore */
    handler->dbuf[y] &= mask[x & 3]; /* reset bits */
    handler->dbuf[y] |= (col << bitshl[x & 3]); /* set new bits */
    x++;
  }
}

static void video_rputpixel_vga(struct video_handler *handler, int x, int y, unsigned char col, int repeat) {
  while (repeat-- > 0) {
    handler->dbuf[((y << 8) + (y << 6)) + x++] = col;
  }
}

void video_rputpixel(struct video_handler *handler, int x, int y, unsigned char col, int repeat) {
  if (handler->mode == 0x13) {
      video_rputpixel_vga(handler, x, y, col, repeat);
    } else if (handler->mode == 0x04) {
      video_rputpixel_cga(handler, x, y, col, repeat);
  }
}

/* render a horizontal line of length len starting at x/y */
void video_hline(struct video_handler *handler, int x, int y, int len, unsigned char color) {
  if (handler->mode == 0x13) { /* optimized for VGA mode 0x13 */
      _fmemset(handler->dbuf + ((y << 8) + (y << 6)) + x, color, len);
    } else { /* not optimized for other modes */
      video_rputpixel_cga(handler, x, y, color, len);
  }
}

/* render a vertical line of length len starting at x/y */
void video_vline(struct video_handler *handler, int x, int y, int len, unsigned char color) {
  while (len > 0) {
    video_rputpixel(handler, x, y++, color, 1);
    len--;
  }
}

/* To draw non-diagonal lines use the much faster hline() and vline() */
void video_line(struct video_handler *handler, int x1, int y1, int x2, int y2, unsigned char color) {
  unsigned l;
  unsigned long x, y;
  int dx, dy;
  x = x1;
  x <<= 9;
  x += 256;
  y = y1;
  y <<= 9;
  y += 256;
  dx = x2 - x1;
  dy = y2 - y1;
  for (l = 0; l < 512; l++) {
    video_rputpixel(handler, (int)(x >> 9), (int)(y >> 9), color, 1);
    x += dx;
    y += dy;
  }
}

void video_rect(struct video_handler *handler, int x, int y, int width, int height, unsigned char color) {
  video_hline(handler, x, y, width, color);                 /*  +--------+  */
  video_vline(handler, x, y, height, color);                /*  |           */
  video_vline(handler, x + width - 1, y, height, color);    /*           |  */
  video_hline(handler, x, y + height - 1, width, color);    /*  +--------+  */
}
