/*
 * This is Mateusz Viste's mode X library. It provides function to deal
 * with the VGA mode X ("tweaked mode") at a resolution of 320x240 pixels,
 * with 256 possible colors and 3 video pages.
 */

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <mem.h>

#ifdef __WATCOMC__
  #include <conio.h>
  #define inportb inp
#endif

/* VGA sequence controller */
#define SC_INDEX      0x03c4
#define SC_DATA       0x03c5

/* VGA CRT controller */
#define CRTC_INDEX    0x03d4
#define CRTC_DATA     0x03d5

unsigned char far *VGAMEM = (unsigned char far*)0xA0000000l;

#define PLANE(x) (1 << (x))


void tweakvga(void) {
  int i;
  int r[][3] = {
        { 0x3c2, 0x0, 0xe3},
        { 0x3d4, 0x0, 0x5f},
        { 0x3d4, 0x1, 0x4f},
        { 0x3d4, 0x2, 0x50},
        { 0x3d4, 0x3, 0x82},
        { 0x3d4, 0x4, 0x54},
        { 0x3d4, 0x5, 0x80},
        { 0x3d4, 0x6, 0xd},
        { 0x3d4, 0x7, 0x3e},
        { 0x3d4, 0x8, 0x0},
        { 0x3d4, 0x9, 0x41},
        { 0x3d4, 0x10, 0xea},
        { 0x3d4, 0x11, 0xac},
        { 0x3d4, 0x12, 0xdf},
        { 0x3d4, 0x13, 0x28},
        { 0x3d4, 0x14, 0x0},
        { 0x3d4, 0x15, 0xe7},
        { 0x3d4, 0x16, 0x6},
        { 0x3d4, 0x17, 0xe3},
        { 0x3c4, 0x1, 0x1},
        { 0x3c4, 0x4, 0x6},
        { 0x3ce, 0x5, 0x40},
        { 0x3ce, 0x6, 0x5},
        { 0x3c0, 0x10, 0x41},
        { 0x3c0, 0x13, 0x0},
        { 0, 0, 0}
  };

  for (i = 0; r[i][0] != 0; i++) {

    switch (r[i][0]) {
                /* First handle special cases: */

                case 0x3c0:
                  inp(0x3da);           /* reset read/write flip-flop */
                  outp(0x3c0, r[i][0] | 0x20);
                  /* ensure VGA output is enabled */
                  outp(0x3c0, r[i][2]);
                  break;

                case 0x3c2:
                case 0x3c3:
                  outp(r[i][0], r[i][2]);   /* directly to the port */
                  break;

                case 0x3c4:
                case 0x3ce:
                case 0x3d4:
                default:                       /* This is the default method: */
                  outp(r[i][0], r[i][1]);      /* select the register index */
                  outp(r[i][0] + 1, r[i][2]);  /* write data */
                  break;
    }
  }
}


/* returns 1 if VGA has been detected, zero otherwise */
int video_detectvga(void) {
  union REGS regs;
  regs.x.ax = 0x1A00; /* can be used to detect the presence of VGA. */
  int86(0x10, &regs, &regs);
  if (regs.h.al == 0x1A) return(1);  /* VGA supported */
  return(0); /* else it's not vga */
}


static void modex_selectplane(int plane) {
  outp(SC_INDEX, 0x02);  /* select the map mask register */
  outp(SC_DATA, plane);  /* write 2^plane */
}


void modex_setpage(int p) {
  unsigned int high_addr, low_addr;
  /* CRT controller registers */
  #define HIGH_ADDRESS 0x0C
  #define LOW_ADDRESS  0x0D

  high_addr = HIGH_ADDRESS | (p & 0xff00);
  low_addr  = LOW_ADDRESS  | (p << 8);

  /* while ((inp(INPUT_STATUS_1) & VRETRACE)); */
  outpw(CRTC_INDEX, high_addr);
  outpw(CRTC_INDEX, low_addr);
  /* while (!(inp(INPUT_STATUS_1) & VRETRACE)); */
}


void putpixel(int x, int y, int c) {
  modex_selectplane(PLANE(x & 3));
  VGAMEM[(y << 6)+(y << 4)+(x >> 2)] = c;
}


void clearscreen(int c) {
  modex_selectplane(15); /* select all 4 planes at once */
  _fmemset(VGAMEM, c, 19700);
}


int main(void) {
  int oldmode;
  int i;
  union REGS regs;

  #define MEMORY_MODE   0x04
  #define UNDERLINE_LOC 0x14
  #define MODE_CONTROL  0x17

  /* first save the current mode, to be able to restore it later */
  regs.h.ah = 0x0F;
  int86(0x10, &regs, &regs);
  oldmode = regs.h.al;

  /* switch to mode 13h */
  regs.h.ah = 0;
  regs.h.al = 0x13;
  int86(0x10, &regs, &regs);

  /* turn off chain-4 mode */
  outp(SC_INDEX, MEMORY_MODE);
  outp(SC_DATA, 0x06);

  /* turn off long mode */
  outp(CRTC_INDEX, UNDERLINE_LOC);
  outp(CRTC_DATA, 0x00);

  /* turn on byte mode */
  outp(CRTC_INDEX, MODE_CONTROL);
  outp(CRTC_DATA, 0xe3);

  tweakvga();

  /* clear the screen here */
  clearscreen(5);

  for (i = 0; i < 242; i++) {
    if (i == 238) continue;
    putpixel(i + 0, i, 1);
    putpixel(i + 1, i, 1);
    putpixel(i + 2, i, 1);
    putpixel(i + 3, i, 15);
    putpixel(i + 4, i, 15);
    putpixel(i + 5, i, 15);
    putpixel(i + 6, i, 4);
    putpixel(i + 7, i, 4);
    putpixel(i + 8, i, 4);
  }

  sleep(10);

  /* restore the initial video mode */
  regs.h.ah = 0;
  regs.h.al = oldmode;
  int86(0x10, &regs, &regs);

  return(0);
}
