/*
 * a few functions for mode 12h programming (640x480 4bpp)
 *
 * This file is part of the Mateusz' DOS Routines (MDR): http://mdr.osdn.io
 * Published under the terms of the MIT License, as stated below.
 *
 * Copyright (C) 2022-2023 Mateusz Viste
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/*
 * Mode 0x12 is a 16 colors (4 bpp) mode, meaning that there are 4 bits used
 * per pixel. It's also a planar mode. VGA has four planes, and for each pixel,
 * each plane holds one bit of the color drawn. Since the information for each
 * pixel is scattered over four memory locations, one needs to switch into each
 * of the 4 planes to set one bit of data in each for the given pixel.
 *
 * Bit 7 of each address contains information about the first pixel, Bit 6 has
 * information about the next pixel, and so on.
 *
 * Plane 0 contains the first bits of all pixels, Plane 1 the second bits and
 * so on.
 *
 * Selecting a plane is performed through the Color Plane Enable Register.
 * It is possible to select 0, 1 or more planes simulteanously, which may be
 * useful if several planes will receive the same information.
 *
 * Planes must be also selected for reading and obviously, only a single plane
 * can be read from at a single time.
 */

#include <string.h>    /* _fmemset() */

#include <mdr\vid12.h> /* include self for control */

#define VGA ((unsigned char far *)(0xA0000000l))
//static unsigned char far *VGA = (void far *)0xA0000000l; /* MK_FP(0xA000, 0); */
static unsigned char vid12_prevmode;


/* enable planes of pmask (0..15) for writing to */
static void vga_plane_sel_wr(unsigned char pmask) {
  _asm {
    /* set write plane */
    mov al, 2
    mov ah, pmask  /* ah is a plane MASK (eg. 3 = planes 0 & 1 enabled) */
    mov dx, 0x3c4
    out dx, ax
  }
}


/* enable plane pid (0..3) for reading from */
static void vga_plane_sel_rd(unsigned char pid) {
  _asm {
    /* set read plane */
    mov al, 4
    mov ah, pid  /* 0..3 (plane id, NOT plane mask!) */
    mov dx, 0x3CE
    out dx, ax
  }
}


/* init video mode 12h (640x480x16c) */
void vid12_init(void) {
  unsigned char prevmode = 0;
  _asm {
    push ax
    push bx
    push cx
    push dx

    /* first save the current mode, to be able to restore it later */
    mov ah, 0x0f /* get current video mode (in al) */
    int 0x10
    mov prevmode, al

    /* now switch to mode 12h */
    mov ax, 0x0012  /* ah=0 "video mode change" / al=0x12 "mode 12h" */
    int 0x10

    /* select "direct write" VGA mode - most probably useless since the
     * int 0x10 BIOS routine is likely to have it set up already, but you
     * never now so let's be sure.
     * Port-Index: 05h              Port: 03ceh
     * usage:      Mode register
     *       d7    Reserved (0)
     *       d6    256-colour mode
     *       d5    Shift register mode
     *       d4    Odd / Even mode
     *       d3    Color compare mode enable (1 = enable)
     *       d2    Reserved (0)
     *       d1,d0 Write mode
     *
     *       d1 d0 Write mode
     *       0  0  Direct write (data rotate, set/reset may apply)
     *       0  1  Use processor latches as write data
     *       1  0  Color plane n (0-3) is filled with the value of
     *             bit n in the write data
     *       1  1  Use (rotated) write data ANDed with Bit mask as
     *             bit mask. Use set/reset as if set/reset was
     *             enable for all planes */
    mov dx, 0x3CE
    mov ax, 0x0005 /* AL = index 5, AH = mode 0 */
    out dx, ax

    pop dx
    pop cx
    pop bx
    pop ax
  }
  /* I am using a local variable because addressing global variables from within
   * asm blocks is not always working (supposedly because of the segment of the
   * variable being different than the segment of this function) */
  vid12_prevmode = prevmode;
}


/* wait until VBLANK */
void vid12_waitvblank(void) {
  _asm {
    mov dx, 0x3DA
    /* wait for "no vblank" (in case there is a vblank right now) */
    WAIT_NOVBLANK:
    in al, dx
    and al, 8
    jnz WAIT_NOVBLANK
    /* wait for "vblank" */
    WAIT_VBLANK:
    in al, dx
    and al, 8
    jz WAIT_VBLANK
  }
}


/* wait until ANY blank: either VBLANK or HBLANK */
void vid12_waitblank(void) {
  _asm {
    mov dx, 0x3DA
    /* wait for "no blank" (in case there is a blank right now) */
    WAIT_NOBLANK:
    in al, dx
    and al, 1
    jnz WAIT_NOBLANK
    /* wait for "vblank" */
    WAIT_BLANK:
    in al, dx
    and al, 1
    jz WAIT_BLANK
  }
}


/* clear screen using color
 * this function is fastest when color is 0 or 15 */
void vid12_cls(unsigned char color) {
  /* reset all bits on all planes (unless color has all bits enabled anyway) */
  if (color != 0xf) {
    vga_plane_sel_wr(15);
    _fmemset(VGA, 0, 38400); /* 640*480/8 = 38400 */

    /* end now if color has no enabled bits */
    if (color == 0) return;
  }

  /* fill selected color planes with set bits */
  vga_plane_sel_wr(color & 15);
  _fmemset(VGA, 0xff, 38400);
}


/* clear a single scanline (0..479) with a solid color (0..15)
 * this function is fastest when color is 0 or 15 */
void vid12_clrline(unsigned short line, unsigned char color) {
  /* convert line index to a memory offset */
  line *= 80;

  /* reset all bits on all planes (unless color has all bits enabled anyway) */
  if (color != 0xf) {
    vga_plane_sel_wr(15);
    _fmemset(VGA + line, 0, 80); /* 640/8 = 80 */

    /* end now if color has no enabled bits */
    if (color == 0) return;
  }

  /* fill selected color planes with set bits */
  vga_plane_sel_wr(color & 15);
  _fmemset(VGA + line, 0xff, 80);
}


/* fill lines from linefirst to linelast with an 8 pixels pattern
 * linelast must be equal to or greater than linelast
 * pattern must be 8 bytes long */
void vid12_linepat(unsigned short linefirst, unsigned short linelast, const unsigned char *pattern) {
  unsigned char byteplane[4]; /* no need to init this, each member will be SHL 8 times anyway */

  /* convert pattern to four byte planes */
  {
    unsigned char i;
    const unsigned char BITDETECT[9] = {0, 1, 1, 1, 1, 1, 1, 1, 1};

    for (i = 0; i < 8; i++) {
      byteplane[0] <<= 1;
      byteplane[0] |= BITDETECT[*pattern & 1];
      byteplane[1] <<= 1;
      byteplane[1] |= BITDETECT[*pattern & 2];
      byteplane[2] <<= 1;
      byteplane[2] |= BITDETECT[*pattern & 4];
      byteplane[3] <<= 1;
      byteplane[3] |= BITDETECT[*pattern & 8];
      pattern++;
    }
  }

  /* convert linelast to a number of bytes from linestart */
  linelast++;
  linelast -= linefirst;
  linelast *= 80;

  /* convert linefirst to a byte offset positions */
  linefirst *= 80;

  /* fill each plane section with its corresponding byte */
  vga_plane_sel_wr(1);
  _fmemset(VGA + linefirst, byteplane[0], linelast);

  vga_plane_sel_wr(2);
  _fmemset(VGA + linefirst, byteplane[1], linelast);

  vga_plane_sel_wr(4);
  _fmemset(VGA + linefirst, byteplane[2], linelast);

  vga_plane_sel_wr(8);
  _fmemset(VGA + linefirst, byteplane[3], linelast);
}


void vid12_close(void) {
  _asm {
    /* restore the initial video mode */
    xor ah, ah
    mov al, vid12_prevmode
    int 0x10
  }
}


void vid12_putpixel(unsigned short x, unsigned short y, unsigned char col) {
  unsigned char pix, npix;
  /* when set as static, these arrays perform 5% faster than when set as const... */
  static unsigned char RESETBIT[8] = {127, 191, 223, 239, 247, 251, 253, 254};
  static unsigned char SETBIT[8] = {128, 64, 32, 16, 8, 4, 2, 1};

  /* translate y into linear pixel offset in VGA memory */
  y *= 80;    /* one 640px line is 80 bytes long (because 4 planes) */
  y += x / 8;

  /* translate x into offset of the pixel within an 8-bytes block of pixels */
  x &= 7;

  vga_plane_sel_rd(0);
  pix = VGA[y];
  if (col & 1) {
    npix = pix | SETBIT[x];
  } else {
    npix = pix & RESETBIT[x];
  }
  if (npix != pix) {
    vga_plane_sel_wr(1);
    VGA[y] = npix;
  }

  vga_plane_sel_rd(1);
  pix = VGA[y];
  if (col & 2) {
    npix = pix | SETBIT[x];
  } else {
    npix = pix & RESETBIT[x];
  }
  if (npix != pix) {
    vga_plane_sel_wr(2);
    VGA[y] = npix;
  }

  vga_plane_sel_rd(2);
  pix = VGA[y];
  if (col & 4) {
    npix = pix | SETBIT[x];
  } else {
    npix = pix & RESETBIT[x];
  }
  if (npix != pix) {
    vga_plane_sel_wr(4);
    VGA[y] = npix;
  }

  vga_plane_sel_rd(3);
  pix = VGA[y];
  if (col & 8) {
    npix = pix | SETBIT[x];
  } else {
    npix = pix & RESETBIT[x];
  }
  if (npix != pix) {
    vga_plane_sel_wr(8);
    VGA[y] = npix;
  }
}


/* draws a horizonatal line from [x1,y] to [x2,y]. x2 must be equal of greater than x1. */
void vid12_hline(unsigned short y, unsigned short x1, unsigned short x2, unsigned char color) {
  unsigned short x, x81, x82;

  /* identify the range of 8-pix blocks to be filled within the 80 bytes of a scanline */
  x81 = (x1 + 7) / 8;
  x82 = x2 / 8;

  /* draw leading pixels unaligned to 8-pixel boundaries */
  for (x = x1; (x < x81 * 8) && (x <= x2); x++) {
    vid12_putpixel(x, y, color);
  }

  /* write full 8-pixels blocks for each plane */
  if (x81 < x82) {
    const unsigned char ff[9] = {0, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
    unsigned char far *lineptr = VGA + (y * 80);
    /* select the proper plane and fill the plane with a suitable amount of 8-pix mono bytes */
    vga_plane_sel_wr(1);
    _fmemset(lineptr + x81, ff[color & 1], x82 - x81);
    vga_plane_sel_wr(2);
    _fmemset(lineptr + x81, ff[color & 2], x82 - x81);
    vga_plane_sel_wr(4);
    _fmemset(lineptr + x81, ff[color & 4], x82 - x81);
    vga_plane_sel_wr(8);
    _fmemset(lineptr + x81, ff[color & 8], x82 - x81);
  }

  /* draw trailing pixels unaligned to 8-pixel boundaries */
  for (x = x82 * 8; (x >= x1) && (x <= x2); x++) {
    vid12_putpixel(x, y, color);
  }
}


/* write an entire scanline (640 pixels) to screen. the scanline is expected
 * to be a serie of 640 bytes having values betwen 0..15 */
void vid12_putscanline(unsigned short scanline, const unsigned char *pixels) {
  unsigned short x;
  unsigned short pix16; /* 16 pixels (single plane of) */
  unsigned short far *scanlineptr;
  unsigned char pmask;

  const unsigned short BITDETECT16[9] = {0, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000};
  const unsigned short BITDETECT15[9] = {0, 0x4000, 0x4000, 0x4000, 0x4000, 0x4000, 0x4000, 0x4000, 0x4000};
  const unsigned short BITDETECT14[9] = {0, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000, 0x2000};
  const unsigned short BITDETECT13[9] = {0, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000, 0x1000};
  const unsigned short BITDETECT12[9] = {0, 0x0800, 0x0800, 0x0800, 0x0800, 0x0800, 0x0800, 0x0800, 0x0800};
  const unsigned short BITDETECT11[9] = {0, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400};
  const unsigned short BITDETECT10[9] = {0, 0x0200, 0x0200, 0x0200, 0x0200, 0x0200, 0x0200, 0x0200, 0x0200};
  const unsigned short BITDETECT09[9] = {0, 0x0100, 0x0100, 0x0100, 0x0100, 0x0100, 0x0100, 0x0100, 0x0100};
  const unsigned short BITDETECT08[9] = {0, 0x0080, 0x0080, 0x0080, 0x0080, 0x0080, 0x0080, 0x0080, 0x0080};
  const unsigned short BITDETECT07[9] = {0, 0x0040, 0x0040, 0x0040, 0x0040, 0x0040, 0x0040, 0x0040, 0x0040};
  const unsigned short BITDETECT06[9] = {0, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020};
  const unsigned short BITDETECT05[9] = {0, 0x0010, 0x0010, 0x0010, 0x0010, 0x0010, 0x0010, 0x0010, 0x0010};
  const unsigned short BITDETECT04[9] = {0, 0x0008, 0x0008, 0x0008, 0x0008, 0x0008, 0x0008, 0x0008, 0x0008};
  const unsigned short BITDETECT03[9] = {0, 0x0004, 0x0004, 0x0004, 0x0004, 0x0004, 0x0004, 0x0004, 0x0004};
  const unsigned short BITDETECT02[9] = {0, 0x0002, 0x0002, 0x0002, 0x0002, 0x0002, 0x0002, 0x0002, 0x0002};
  const unsigned short BITDETECT01[9] = {0, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001, 0x0001};

  scanline *= 80; /* translate scanline number into memory offset */

  /* write each plane */
  for (pmask = 8; pmask != 0; pmask >>= 1) { /* 8, 4, 2, 1 */
    scanlineptr = (unsigned short far *)(VGA + scanline);

    /* select plane n */
    vga_plane_sel_wr(pmask);

    for (x = 0; x < 625; x += 16) {

      /* pack the Nth bit of the next 16 pixels into a single word */
      pix16 = BITDETECT16[pixels[x+8] & pmask];
      pix16 |= BITDETECT15[pixels[x+9] & pmask];
      pix16 |= BITDETECT14[pixels[x+10] & pmask];
      pix16 |= BITDETECT13[pixels[x+11] & pmask];
      pix16 |= BITDETECT12[pixels[x+12] & pmask];
      pix16 |= BITDETECT11[pixels[x+13] & pmask];
      pix16 |= BITDETECT10[pixels[x+14] & pmask];
      pix16 |= BITDETECT09[pixels[x+15] & pmask];
      pix16 |= BITDETECT08[pixels[x] & pmask];
      pix16 |= BITDETECT07[pixels[x+1] & pmask];
      pix16 |= BITDETECT06[pixels[x+2] & pmask];
      pix16 |= BITDETECT05[pixels[x+3] & pmask];
      pix16 |= BITDETECT04[pixels[x+4] & pmask];
      pix16 |= BITDETECT03[pixels[x+5] & pmask];
      pix16 |= BITDETECT02[pixels[x+6] & pmask];
      pix16 |= BITDETECT01[pixels[x+7] & pmask];

      *scanlineptr = pix16;
      scanlineptr++;
    }

  }
}


void vid12_setpalette(unsigned char index, unsigned char r, unsigned char g, unsigned char b) {
  /* The default mapping of color indexes in VGA 4 bpp mode is weird:
   *
   * 0  1  2  3  4  5  6 7  8  9  A  B  C  D  E  F     palette register
   * 0  1  2  3  4  5 14 7 38 39 3A 3B 3C 3D 3E 3F     color register
   *
   * This can be changed through the VGA register 0x3c0, index 00..0fh, but
   * I found out such remapping does not work reliably on all hardware, hence
   * I prefer to remap it here, in software. It's also much simpler.
   */
  switch (index) {
    case 0x06:
      index = 0x14;
      break;
    case 0x08:
    case 0x09:
    case 0x0A:
    case 0x0B:
    case 0x0C:
    case 0x0D:
    case 0x0E:
    case 0x0F:
      index += 0x30;
      break;
  }
  _asm {
    mov al, index
    mov dx, 0x3C8 /* palette register write index */
    out dx, al
    inc dx  /* now I will write to 0x3C9 (palette rgb values register) */
    /* red */
    mov al, r
    out dx, al
    /* green */
    mov al, g
    out dx, al
    /* blue */
    mov al, b
    out dx, al
  }
}
