/*
 * WAVE-loading routines
 *
 * Copyright (C) 2022 Mateusz Viste
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#include <stdio.h>  /* FILE and friends */
#include <string.h> /* bzero(), bcmp() */

#include <mdr\wave.h>


int wave_anal(struct wave_t *wav, FILE *fd) {
  /* a WAVE file is basically a RIFF container that contains a WAVE chunk:
   * rID   (4 bytes)    "RIFF"
   * rLEN  (4 bytes)    length of the data that follows
   * wID   (4 bytes)    "WAVE"
   * FMT   (20 bytes)   a "format" chunk
   * DATA  (variable)   a "data" chunk that contains actual samples
   *
   * There may possibly be other "chunks" after wID. These can probably be
   * ignored as they are likely to contain only boring metdata.
   *
   * the FMT chunk is:
   * fID   (4 bytes)    "fmt " (note the trailing space)
   * fLEN  (4 bytes)    length of the fmt chunk (at least 16 bytes)
   * wFMT  (2 bytes)    audio format: 1=PCM 3=IEEE_float 6=8bit_ALAW 7=8bit_ULAW
   * nCH   (2 bytes)    number of channels
   * nSMP  (4 bytes)    sampling rate (blocks per second)
   * nAVG  (4 bytes)    data rate
   * ALIGN (2 bytes)    data block size (aligned bytes)
   * wBPP  (2 bytes)    bits per samples
   * (...)
   *
   * the DATA chunk is:
   * wID   (4 bytes)    "data"
   * wLEN  (4 bytes)    length of data that follows
   */

  unsigned long chunklen, rifflen, bytesread = 0;
  unsigned char buf[16];

  /* zero out the wav struct */
  bzero(wav, sizeof(*wav));

  /* read the riff header + WAVE id */
  if (fread(buf, 1, 12, fd) != 12) return(-1);
  if (bcmp(buf, "RIFF", 4) != 0) return(-2);
  if (bcmp(buf + 8, "WAVE", 4) != 0) return(-3);

  rifflen = ((unsigned long *)(buf))[1];
  if (rifflen < 28) return(-4);
  bytesread = 4; /* "WAVE" */

  /* read chunks until a FMT chunk is found */
  for (;;) {
    if (fread(buf, 1, 8, fd) != 8) return(-5);
    bytesread += 8;
    chunklen = ((unsigned long *)(buf))[1];
    if (bcmp(buf, "fmt ", 4) == 0) break;
    /* skip the chunk */
    if (fseek(fd, chunklen, SEEK_CUR) != 0) return(-6);
    bytesread += chunklen;
    if (bytesread >= rifflen) return(-7);
  }

  /* found the fmt chunk */
  if (chunklen < 16) return(-8);
  if (fread(buf, 1, 16, fd) != 16) return(-9);
  bytesread += 16;
  wav->format = buf[1];
  wav->format <<= 8;
  wav->format |= buf[0];
  wav->channels = buf[3];
  wav->channels <<= 8;
  wav->channels |= buf[2];
  wav->rate = buf[7];
  wav->rate <<= 8;
  wav->rate |= buf[6];
  wav->rate <<= 8;
  wav->rate |= buf[5];
  wav->rate <<= 8;
  wav->rate |= buf[4];
  wav->bitdepth = buf[15];
  wav->bitdepth <<= 8;
  wav->bitdepth |= buf[14];

  /* read chunks until a WAV chunk is found */
  for (;;) {
    if (fread(buf, 1, 8, fd) != 8) return(-5);
    bytesread += 8;
    chunklen = ((unsigned long *)(buf))[1];
    if (bcmp(buf, "data", 4) == 0) break;
    /* skip the chunk */
    if (fseek(fd, chunklen, SEEK_CUR) != 0) return(-6);
    bytesread += chunklen;
    if (bytesread >= rifflen) return(-7);
  }

  /* fill in offset and length into the wave_t struct and return success */
  wav->dataoffset = bytesread + 8;  /* +8 accounts for the RIFF header */
  wav->len = chunklen;

  return(0);
}
